@@ -30,10 +30,12 @@
 
 G_BEGIN_DECLS
 
-#define VTE_SPAWN_NO_PARENT_ENVV (1 << 25)
+#define VTE_SPAWN_NO_PARENT_ENVV        (1 << 25)
+#define VTE_SPAWN_NO_SYSTEMD_SCOPE      (1 << 26)
+#define VTE_SPAWN_REQUIRE_SYSTEMD_SCOPE (1 << 27)
 
 _VTE_PUBLIC
-GQuark vte_pty_error_quark (void);
+GQuark vte_pty_error_quark (void) _VTE_CXX_NOEXCEPT;
 
 /**
  * VTE_PTY_ERROR:
@@ -61,35 +63,35 @@ GType vte_pty_get_type (void);
 _VTE_PUBLIC
 VtePty *vte_pty_new_sync (VtePtyFlags flags,
                           GCancellable *cancellable,
-                          GError **error);
+                          GError **error) _VTE_CXX_NOEXCEPT;
 
 _VTE_PUBLIC
 VtePty *vte_pty_new_foreign_sync (int fd,
                                   GCancellable *cancellable,
-                                  GError **error);
+                                  GError **error) _VTE_CXX_NOEXCEPT;
 
 _VTE_PUBLIC
-int vte_pty_get_fd (VtePty *pty) _VTE_GNUC_NONNULL(1);
+int vte_pty_get_fd (VtePty *pty) _VTE_CXX_NOEXCEPT _VTE_GNUC_NONNULL(1);
 
 _VTE_PUBLIC
-void vte_pty_child_setup (VtePty *pty) _VTE_GNUC_NONNULL(1);
+void vte_pty_child_setup (VtePty *pty) _VTE_CXX_NOEXCEPT _VTE_GNUC_NONNULL(1);
 
 _VTE_PUBLIC
 gboolean vte_pty_get_size (VtePty *pty,
                            int *rows,
                            int *columns,
-                           GError **error) _VTE_GNUC_NONNULL(1);
+                           GError **error) _VTE_CXX_NOEXCEPT _VTE_GNUC_NONNULL(1);
 
 _VTE_PUBLIC
 gboolean vte_pty_set_size (VtePty *pty,
                            int rows,
                            int columns,
-                           GError **error) _VTE_GNUC_NONNULL(1);
+                           GError **error) _VTE_CXX_NOEXCEPT _VTE_GNUC_NONNULL(1);
 
 _VTE_PUBLIC
 gboolean vte_pty_set_utf8 (VtePty *pty,
                            gboolean utf8,
-                           GError **error) _VTE_GNUC_NONNULL(1);
+                           GError **error) _VTE_CXX_NOEXCEPT _VTE_GNUC_NONNULL(1);
 
 G_DEFINE_AUTOPTR_CLEANUP_FUNC(VtePty, g_object_unref)
 
@@ -105,13 +107,30 @@ void vte_pty_spawn_async(VtePty *pty,
                          int timeout,
                          GCancellable *cancellable,
                          GAsyncReadyCallback callback,
-                         gpointer user_data) _VTE_GNUC_NONNULL(1) _VTE_GNUC_NONNULL(3);
+                         gpointer user_data) _VTE_CXX_NOEXCEPT _VTE_GNUC_NONNULL(1) _VTE_GNUC_NONNULL(3);
 
 _VTE_PUBLIC
+void vte_pty_spawn_with_fds_async(VtePty *pty,
+                                  char const* working_directory,
+                                  char const* const* argv,
+                                  char const* const* envv,
+                                  int const* fds,
+                                  int n_fds,
+                                  int const* map_fds,
+                                  int n_map_fds,
+                                  GSpawnFlags spawn_flags,
+                                  GSpawnChildSetupFunc child_setup,
+                                  gpointer child_setup_data,
+                                  GDestroyNotify child_setup_data_destroy,
+                                  int timeout,
+                                  GCancellable *cancellable,
+                                  GAsyncReadyCallback callback,
+                                  gpointer user_data) _VTE_CXX_NOEXCEPT _VTE_GNUC_NONNULL(1) _VTE_GNUC_NONNULL(3);
+
 gboolean vte_pty_spawn_finish(VtePty *pty,
                               GAsyncResult *result,
                               GPid *child_pid /* out */,
-                              GError **error) _VTE_GNUC_NONNULL(1) _VTE_GNUC_NONNULL(2);
+                              GError **error) _VTE_CXX_NOEXCEPT _VTE_GNUC_NONNULL(1) _VTE_GNUC_NONNULL(2);
 
 G_END_DECLS
 
